{
  :: This unit contains a wrapper implementation for the Messenger Plus!
  :: Plugin SDK designed to convert a lineair C++-style approach to a
  :: more elegant Delphi-style interface.

  :$
  :$
  :$ This unit is released under the zlib/libpng OSI-approved license.
  :$ For more information: http://www.opensource.org/
  :$ /n/n
  :$ /n/n
  :$ Copyright (c) 2004 X2Software
  :$ /n/n
  :$ This software is provided 'as-is', without any express or implied warranty.
  :$ In no event will the authors be held liable for any damages arising from
  :$ the use of this software.
  :$ /n/n
  :$ Permission is granted to anyone to use this software for any purpose,
  :$ including commercial applications, and to alter it and redistribute it
  :$ freely, subject to the following restrictions:
  :$ /n/n
  :$ 1. The origin of this software must not be misrepresented; you must not
  :$ claim that you wrote the original software. If you use this software in a
  :$ product, an acknowledgment in the product documentation would be
  :$ appreciated but is not required.
  :$ /n/n
  :$ 2. Altered source versions must be plainly marked as such, and must not be
  :$ misrepresented as being the original software.
  :$ /n/n
  :$ 3. This notice may not be removed or altered from any source distribution.
}
unit MPPlugin;

interface
uses
  SysUtils,
  Classes,
  Graphics,
  MPPluginHeader;

type
  //:$ Raised when you either pass TMPPlugin as a class in RegisterPlugin
  //:$ or pass a class not descending from TMPPlugin
  EMPInvalidClass = class(Exception);

  //:$ Contains the information for command and tag callback functions
  TMPCommandInfo  = record
    Command:        String;     // The name of the command
    Params:         String;     // The parameters passed to the command
                                // (will be empty for tags, contains the
                                // remaining text of the message for
                                // notifications)
    ContactName:    String;     // Name of the contact which triggered the
                                // notification (not available for commands
                                // and tags)
    Conversation:   IDispatch;  // The IMessengerConversationWnd interface.
                                // Not always available.
  end;

  //:$ Contains the interfaces filters may use
  TMPFilterInfo   = record
    Messenger:      IDispatch;
    Conversation:   IDispatch;
  end;

  //:$ Prototype for command or tag callback functions
  TMPCommandProc  = function(const AInfo: TMPCommandInfo): String of object;

  //:$ Prototype for filter callback functions
  TMPFilterProc   = procedure(var ASource: String;
                              const AInfo: TMPFilterInfo);

  {
    :$ Contains known Messenger Plus! versions
  }
  TMPVersion      = (mp220, mp221, mp250, mp300);

  {
    :$ Holds information about a single command or tag

    :: Name corresponds to the name showed in the Plus! plugin menu. Command
    :: is the command or tag which triggers this action. Help is only used
    :: for commands accepting parameters, to provide the user with a short
    :: help string to indicate they can enter the parameters there. Callback
    :: references the callback function which is called when the command is
    :: executed.
  }
  TMPCommand      = class(TCollectionItem)
  private
    FName:          String;
    FCommand:       String;
    FHelp:          String;
    FCallback:      TMPCommandProc;
  public
    property Name:        String          read FName      write FName;
    property Command:     String          read FCommand   write FCommand;
    property Help:        String          read FHelp      write FHelp;
    property Callback:    TMPCommandProc  read FCallback  write FCallback;
  end;

  {
    :$ Holds a collection of commands or tags
  }
  TMPCommands     = class(TCollection)
  private
    function GetItem(Index: Integer): TMPCommand;
    procedure SetItem(Index: Integer; const Value: TMPCommand);
  public
    constructor Create();

    //:$ Adds a command or tag to the list
    function Add(): TMPCommand; overload;

    //:$ Adds a command or tag to the list
    function Add(const ACommand, AName: String;
                 const ACallback: TMPCommandProc;
                 const AHelp: String = ''): TMPCommand; overload;

    //:$ Finds a command
    function Find(const ACommand: String): TMPCommand;

    property Items[Index: Integer]: TMPCommand  read GetItem
                                                write SetItem; default;
  end;


  {
    :$ Holds information about a notification

    :: Command corresponds to the 5-character notification command. Callback
    :: references the callback function which is called when the command is
    :: executed.
  }
  TMPNotify         = class(TCollectionItem)
  private
    FCommand:       String;
    FCallback:      TMPCommandProc;
  public
    property Command:     String          read FCommand   write FCommand;
    property Callback:    TMPCommandProc  read FCallback  write FCallback;
  end;

  {
    :$ Holds a collection of notifications
  }
  TMPNotifications  = class(TCollection)
  private
    function GetItem(Index: Integer): TMPNotify;
    procedure SetItem(Index: Integer; const Value: TMPNotify);
  public
    constructor Create();

    //:$ Adds a notification to the list
    function Add(): TMPNotify; overload;

    //:$ Adds a notification to the list
    function Add(const ACommand: String;
                 const ACallback: TMPCommandProc): TMPNotify; overload;

    //:$ Finds a notification
    function Find(const ACommand: String): TMPNotify;

    property Items[Index: Integer]: TMPNotify read GetItem
                                              write SetItem; default;
  end;


  {
    :$ Base Messenger Plus! plugin

    :: Use this class as a base class when writing Plus! plugins. After defining
    :: a descendant, use the RegisterPlugin function to register your class.
    :: Override the Initialize function to detect Plus! initializing your
    :: plugin, the Version and UserEmail properties will be set at that point.
    :: It is also wise to set the PluginName property in that function.
    :: /n/n
    :: Registering tags and commands is relatively easy, but might seem some
    :: explanation. To register a command, call Commands.Add (or Tags.Add) in
    :: the Initialize function. You pass it the command, callback function and
    :: help string. The command refers to /x<command>, the help string will
    :: be displayed in Plus!'s plugin menu. The callback function must be a
    :: method of your TMPPlugin descendant and conform to the following
    :: prototype:
    :: /n/n
    :: function MyCommand(const AInfo: TMPCommandInfo): String;
  }
  TMPPlugin       = class(TObject)
  private
    FPluginName:    String;
    FMessenger:     IDispatch;
    FUserEmail:     String;
    FVersion:       TMPVersion;

    FCommands:      TMPCommands;
    FTags:          TMPCommands;
    FNotify:        TMPNotifications;
  protected
    function ExecuteCommand(const ACommand, ACommandArg: String;
                            const AParam: TPluginParam;
                            out AResult: String): Boolean; virtual;
    function ExecuteTag(const ATag: String; const AParam: TPluginParam;
                        out AResult: String): Boolean; virtual;
    function ExecuteNotify(const ACommand: String;
                           const AText: String;
                           const AParam: TPluginParam;
                           out AResult: String): Boolean; virtual;
  public
    constructor Create();
    destructor Destroy(); override;

    //:$ Called when Plus! initializes the plugin
    //:: Overwrite Initialize instead of using a constructor to get
    //:: access to all the properties and define the tags and commands
    //:: for the plugin.
    procedure Initialize(); virtual; abstract;

    //:$ Configures the plugin
    //:: Called when a user clicks the Configure button in the Plus! plugins
    //:: dialog. Only used when the Plus! version is at least 3.0.
    //:! Due to the design of the plugin framework the Configure button will
    //:! always be enabled, therefore you are strongly encouraged to override
    //:! this function and provide a configuration for your plugin.
    //:! When dealing with older Plus! versions it is best to register
    //:! a configuration command which calls Configure().
    procedure Configure(); virtual;


    //:$ Contains a list of all registered commands
    property Commands:    TMPCommands read FCommands;

    //:$ Contains a list of all registered tags
    property Tags:        TMPCommands read FTags;

    //:$ Contains a list of all registered notifications
    property Notify:      TMPNotifications  read FNotify;


    //:$ Contains a reference to the Messenger API interface
    //:! The interface may be nil, be sure to check for it!
    property Messenger:   IDispatch   read FMessenger   write FMessenger;

    //:$ Contains the plugin name, set this in the constructor so
    //:$ Messenger Plus! can use it in it's plugin menu
    property PluginName:  String      read FPluginName  write FPluginName;

    //:$ Contains the UserEmail property as passed by Messenger Plus!
    property UserEmail:   String      read FUserEmail   write FUserEmail;

    //:$ Contains the Messenger Plus! version
    property Version:     TMPVersion  read FVersion     write FVersion;
  end;
  TMPPluginClass  = class of TMPPlugin;


  //:$ Encloses the specified string in Plus! Bold tags
  function MPBold(const ASource: String): String;

  //:$ Encloses the specified string in Plus! Italic tags
  function MPItalic(const ASource: String): String;

  //:$ Encloses the specified string in Plus! Underline tags
  function MPUnderline(const ASource: String): String;

  //:$ Encloses the specified string in Plus! Strikeout tags
  function MPStrikeout(const ASource: String): String;

  //:$ Encloses the specified string in Plus! noicon tags
  //:! This function can only be used on the full or beginning of a message!
  function MPNoIcon(const ASource: String): String;

  //:$ Encloses the specified string in Plus! action tags (/me)
  //:! This function can only be used on the full or beginning of a message!
  function MPAction(const ASource: String): String;

  //:$ Encloses the specified string in Plus! action tags (/me)
  //:: The command must be 5 characters long, will be cut if too long, or
  //:: padded with spaces if too short!
  //:! This function can only be used on the full or beginning of a message!
  function MPNotify(const ACommand: String; ASource: String): String;

  //:$ Encloses the specified string in Plus! color tags
  function MPColor(const ASource: String; const AColor: TColor): String; overload;

  //:$ Encloses the specified string in Plus! color tags using a
  //:$ background color
  function MPColor(const ASource: String; const AColor: TColor;
                   const ABackColor: TColor): String; overload;

  //:$ Displays a Messenger popup (also known as 'Toast')
  procedure MPPopup(const AMessage: String;
                    const ATitle: String = ''; AProgram: String = '';
                    const APlaySound: Boolean = True);


  //:$ Simple wrapper around StrLCopy
  procedure LCopy(ADest: PChar; const ASource: String;
                  const AMaxLen: Integer);


  //:$ Registers a Messenger Plus! plugin
  //:: Call this function in your DLL's project source.
  procedure RegisterPlugin(const AClass: TMPPluginClass);

  //:$ Registers a filter
  //:: Output generated by the plugin is processed by registered filters
  //:: before being passed back to Plus!. A common example of this
  //:: is to parse tags in the output not processed by Plus!.
  procedure RegisterFilter(const ACallback: TMPFilterProc);


  //:$ Runs the registered filters on the specified source
  procedure RunFilters(var ASource: String;
                       const AInfo: TMPFilterInfo);

const
  clOrange  = TColor($007FFC);

implementation
uses
  Windows;

var
  GPlugin:        TMPPluginClass  = nil;
  GPluginInst:    TMPPlugin       = nil;
  GFilters:       array of TMPFilterProc;


{============================ TMPCommands
  Item Management
========================================}
constructor TMPCommands.Create;
begin
  inherited Create(TMPCommand);
end;


function TMPCommands.Add(): TMPCommand;
begin
  Result  := TMPCommand(inherited Add());
end;

function TMPCommands.Add(const ACommand, AName: String;
                         const ACallback: TMPCommandProc;
                         const AHelp: String): TMPCommand;
begin
  Result          := Add();
  Result.Command  := ACommand;
  Result.Name     := AName;
  Result.Help     := AHelp;
  Result.Callback := ACallback;
end;

function TMPCommands.Find;
var
  iCommand:       Integer;

begin
  Result  := nil;
  for iCommand  := Count - 1 downto 0 do
    if Items[iCommand].Command = ACommand then begin
      Result  := Items[iCommand];
      break;
    end;
end;


function TMPCommands.GetItem;
begin
  Result  := TMPCommand(inherited GetItem(Index));
end;

procedure TMPCommands.SetItem;
begin
  inherited SetItem(Index, Value);
end;


{======================= TMPNotifications
  Item Management
========================================}
constructor TMPNotifications.Create;
begin
  inherited Create(TMPNotify);
end;


function TMPNotifications.Add(const ACommand: String;
                              const ACallback: TMPCommandProc): TMPNotify;
begin
  Result          := Add();
  Result.Command  := ACommand;
  Result.Callback := ACallback;
end;

function TMPNotifications.Add(): TMPNotify;
begin
  Result  := TMPNotify(inherited Add());
end;


function TMPNotifications.Find;
var
  iCommand:       Integer;

begin
  Result  := nil;
  for iCommand  := Count - 1 downto 0 do
    if Items[iCommand].Command = ACommand then begin
      Result  := Items[iCommand];
      break;
    end;
end;

function TMPNotifications.GetItem;
begin
  Result  := TMPNotify(inherited GetItem(Index));
end;

procedure TMPNotifications.SetItem;
begin
  inherited SetItem(Index, Value);
end;


{============================== TMPPlugin
  Initialization
========================================}
constructor TMPPlugin.Create;
begin
  inherited;

  FPluginName := '<unnamed plugin>';
  FCommands   := TMPCommands.Create();
  FTags       := TMPCommands.Create();
  FNotify     := TMPNotifications.Create();
end;

destructor TMPPlugin.Destroy;
begin
  FreeAndNil(FNotify);
  FreeAndNil(FTags);
  FreeAndNil(FCommands);

  inherited;
end;


procedure TMPPlugin.Configure;
begin
  MessageBox(0, 'This plugin can not be configured.', 'Sorry', MB_OK or
             MB_ICONINFORMATION);
end;


{============================== TMPPlugin
  Execution
========================================}
function TMPPlugin.ExecuteCommand;
var
  pCommand:       TMPCommand;
  pInfo:          TMPCommandInfo;

begin
  Result    := False;
  pCommand  := Commands.Find(ACommand);
  if (Assigned(pCommand)) and (Assigned(pCommand.Callback)) then begin
    pInfo.Command       := ACommand;
    pInfo.Params        := ACommandArg;
    pInfo.ContactName   := '';
    pInfo.Conversation  := AParam.iConversationWnd; 
    AResult             := pCommand.Callback(pInfo);
    Result              := True;
  end;
end;

function TMPPlugin.ExecuteTag;
var
  pCommand:       TMPCommand;
  pInfo:          TMPCommandInfo;

begin
  Result    := False;
  pCommand  := Tags.Find(ATag);
  if (Assigned(pCommand)) and (Assigned(pCommand.Callback)) then begin
    pInfo.Command       := ATag;
    pInfo.Params        := '';
    pInfo.ContactName   := '';
    pInfo.Conversation  := AParam.iConversationWnd;
    AResult             := pCommand.Callback(pInfo);
    Result              := True;
  end;
end;

function TMPPlugin.ExecuteNotify;
var
  pCommand:     TMPNotify;
  pInfo:          TMPCommandInfo;

begin
  Result    := False;
  pCommand  := Notify.Find(ACommand);
  if (Assigned(pCommand)) and (Assigned(pCommand.Callback)) then begin
    pInfo.Command       := ACommand;
    pInfo.Params        := AText;
    pInfo.ContactName   := AParam.sContactName;
    pInfo.Conversation  := AParam.iConversationWnd;
    AResult             := pCommand.Callback(pInfo);
    Result              := True;
  end;
end;


{========================================
  Plugin Registration
========================================}
procedure RegisterPlugin;
var
  bRestore:     Boolean;
  sUserEmail:   String;
  pVersion:     TMPVersion;
  ifMessenger:  IDispatch;

begin
  sUserEmail  := '';
  pVersion    := mp220;

  // If a plugin is already active, we need to make sure our new plugin
  // gets initialized instead. Normally you'd only call RegisterPlugin in the
  // project source, but I like to account for abnormal situations in advance.
  bRestore  := Assigned(GPluginInst);
  if bRestore then begin
    ifMessenger := GPluginInst.Messenger;
    sUserEmail  := GPluginInst.UserEmail;
    pVersion    := GPluginInst.Version;
    FreeAndNil(GPluginInst);
  end;

  GPlugin := AClass;

  if bRestore then begin
    GPluginInst           := GPlugin.Create();
    GPluginInst.Messenger := ifMessenger;
    GPluginInst.UserEmail := sUserEmail;
    GPluginInst.Version   := pVersion;
    GPluginInst.Initialize();
  end;
end;


{========================================
  Filters
========================================}
procedure RegisterFilter;
var
  iIndex:       Integer;

begin
  iIndex            := High(GFilters) + 1;
  SetLength(GFilters, iIndex + 1);
  GFilters[iIndex]  := ACallback;
end;

procedure RunFilters;
var
  iProc:      Integer;

begin
  for iProc := 0 to High(GFilters) do
    GFilters[iProc](ASource, AInfo);
end;


{========================================
  Messenger Plus! Format Wrappers
========================================}
function ColorToPlus(const AColor: TColor): String;
var
  sColor:       String;
  iColor:       Cardinal;

begin
  case AColor of
    // Check for predefined colors
    clWhite:      sColor  := '0';
    clBlack:      sColor  := '1';
    clNavy:       sColor  := '2';
    clGreen:      sColor  := '3';
    clRed:        sColor  := '4';
    clMaroon:     sColor  := '5';
    clPurple:     sColor  := '6';
    clOrange:     sColor  := '7';
    clYellow:     sColor  := '8';
    clLime:       sColor  := '9';
    clTeal:       sColor  := '10';
    clAqua:       sColor  := '11';
    clBlue:       sColor  := '12';
    clFuchsia:    sColor  := '13';
    clGray:       sColor  := '14';
    clSilver:     sColor  := '15';
  else
    // Use RGB values
    iColor  := ColorToRGB(AColor);
    sColor  := Format('(%.3d,%.3d,%.3d)', [GetRValue(iColor),
                                     GetGValue(iColor),
                                     GetBValue(iColor)]);
  end;

  Result  := sColor;
end;


function MPBold;
begin
  Result  := sCCBold + ASource + sCCBold;
end;

function MPItalic;
begin
  Result  := sCCItalic + ASource + sCCItalic;
end;

function MPUnderline;
begin
  Result  := sCCUnderline + ASource + sCCUnderline;
end;

function MPStrikeout;
begin
  Result  := sCCStrikeout + ASource + sCCStrikeout;
end;

function MPNoIcon;
begin
  Result  := sCCNoIcon + ASource;
end;

function MPAction;
begin
  Result  := sCCMeStart + ASource + sCCMeEnd;
end;

function MPNotify;
var
  sNotify:      String;

begin
  sNotify := ACommand;
  if Length(sNotify) < 5 then
    sNotify := sNotify + StringOfChar(' ', 5 - Length(sNotify));

  Result  := sCCNotify + sNotify + ASource;
end;

function MPColor(const ASource: String; const AColor: TColor): String;
begin
  Result  := sCCColor + ColorToPlus(AColor) + ASource + sCCColor;
end;

function MPColor(const ASource: String; const AColor: TColor;
                 const ABackColor: TColor): String;
begin
  Result  := sCCColor + ColorToPlus(AColor) + ',' +
             ColorToPlus(ABackColor) + ASource + sCCColor;
end;


procedure MPPopup;
var
  pTitle:       PChar;
  pProgram:     PChar;

begin
  pTitle    := nil;
  pProgram  := nil;

  if Length(ATitle) > 0 then
    pTitle    := PChar(ATitle);

  if Length(AProgram) > 0 then
    pProgram  := PChar(AProgram);

  DisplayToast(PChar(AMessage), pTitle, pProgram, APlaySound);
end;

procedure LCopy;
begin
  StrLCopy(ADest, PChar(ASource), AMaxLen - 1);
end;


{========================================
  Messenger Plus! Plugin SDK
========================================}
function Initialize(nVersion: DWORD; const sUserEmail: PChar;
                    iMessengerObj: IDispatch): BOOL; stdcall;
begin
  Result  := False;
  if not Assigned(GPlugin) then
    exit;

  FreeAndNil(GPluginInst);
  GPluginInst           := GPlugin.Create();
  GPluginInst.Messenger := iMessengerObj;
  GPluginInst.UserEmail := sUserEmail;
  GPluginInst.Version   := TMPVersion(nVersion);
  GPluginInst.Initialize();
  Result                := True;
end;

procedure Uninitialize(); stdcall;
begin
  FreeAndNil(GPluginInst);
end;

procedure Configure(); stdcall;
begin
  if not Assigned(GPluginInst) then
    exit;

  GPluginInst.Configure();
end;

function PublishInfo(const sPluginName: PChar;
                     const aCommands, aTags: PPluginPublishList): BOOL; stdcall;
var
  iItem:      Integer;
  iCount:     Integer;

begin
  Result  := False;
  if not Assigned(GPluginInst) then
    exit;

  LCopy(sPluginName, GPluginInst.PluginName, 128);

  // Get commands
  iCount  := GPluginInst.Commands.Count;
  if iCount > 50 then
    iCount  := 50;

  aCommands^.nCount := iCount;
  for iItem := 0 to iCount - 1 do begin
    LCopy(aCommands^.sName[iItem], GPluginInst.Commands[iItem].Name, 128);
    LCopy(aCommands^.sValue[iItem], GPluginInst.Commands[iItem].Command, 128);
    LCopy(aCommands^.sHelp[iItem], GPluginInst.Commands[iItem].Help, 128);
  end;

  // Get tags
  iCount  := GPluginInst.Tags.Count;
  if iCount > 50 then
    iCount  := 50;

  aTags^.nCount     := iCount;
  for iItem := 0 to iCount - 1 do begin
    LCopy(aTags^.sName[iItem], GPluginInst.Tags[iItem].Name, 128);
    LCopy(aTags^.sValue[iItem], '(!' + GPluginInst.Tags[iItem].Command + ')', 128);
    LCopy(aTags^.sHelp[iItem], GPluginInst.Tags[iItem].Help, 128);
  end;

  Result  := True;
end;

function ParseCommand(const sCommand, sCommandArg: PChar;
                      const pParam: PPluginParam; sResult: PChar): BOOL; stdcall;
var
  sCmd:       String;
  sTemp:      String;
  pInfo:      TMPFilterInfo;

begin
  Result  := False;
  if not Assigned(GPluginInst) then
    exit;

  // Cut away the /
  sCmd    := sCommand;
  Delete(sCmd, 1, 1);

  Result  := GPluginInst.ExecuteCommand(sCmd, sCommandArg, pParam^, sTemp);

  if Result then begin
    pInfo.Messenger     := GPluginInst.Messenger;
    pInfo.Conversation  := pParam^.iConversationWnd;
    RunFilters(sTemp, pInfo);
    LCopy(sResult, sTemp, 1100);
  end;
end;

function ParseTag(const sTag: PChar; const pParam: PPluginParam;
                  sResult: PChar): BOOL; stdcall;
var
  sCmd:       String;
  sTemp:      String;
  pInfo:      TMPFilterInfo;

begin
  Result  := False;
  if not Assigned(GPluginInst) then
    exit;

  // Cut away the (! and ) parts
  sCmd    := sTag;
  Delete(sCmd, 1, 2);
  SetLength(sCmd, Length(sCmd) - 1);

  Result  := GPluginInst.ExecuteTag(sCmd, pParam^, sTemp);

  if Result then begin
    pInfo.Messenger     := GPluginInst.Messenger;
    pInfo.Conversation  := pParam^.iConversationWnd;
    RunFilters(sTemp, pInfo);
    LCopy(sResult, sTemp, 1100);
  end;
end;

function ReceiveNotify(const sNotifyCode, sText: PChar;
                       const pParam: PPluginParam;
                       sTextToSend: PChar): BOOL; stdcall;
var
  sTemp:      String;
  pInfo:      TMPFilterInfo;

begin
  Result  := False;
  if not Assigned(GPluginInst) then
    exit;

  Result  := GPluginInst.ExecuteNotify(sNotifyCode, sText, pParam^, sTemp);

  if Result then begin
    pInfo.Messenger     := GPluginInst.Messenger;
    pInfo.Conversation  := pParam^.iConversationWnd;
    RunFilters(sTemp, pInfo);
    LCopy(sTextToSend, sTemp, 1100);
  end;
end;


exports
  Initialize,
  Uninitialize,
  Configure,
  PublishInfo,
  ParseCommand,
  ParseTag,
  ReceiveNotify;

end.
